import sys

from pydantic import PostgresDsn, MySQLDsn
from pydantic_core import Url
from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker
from .errors import ServerError

postgres_allowed_schemes = [
    'postgres',
    'postgresql',
    'postgresql+asyncpg',
    'postgresql+pg8000',
    'postgresql+psycopg',
    'postgresql+psycopg2',
    'postgresql+psycopg2cffi',
    'postgresql+py-postgresql',
    'postgresql+pygresql',
]

mysql_allowed_schemes = [
    'mysql',
    'mysql+mysqlconnector',
    'mysql+aiomysql',
    'mysql+asyncmy',
    'mysql+mysqldb',
    'mysql+pymysql',
    'mysql+cymysql',
    'mysql+pyodbc',
]


def validate_database_dns(url: str):
    url = Url(url)
    if url.scheme in postgres_allowed_schemes:
        dsn = PostgresDsn.build(
            scheme='postgresql+asyncpg',
            username=url.username,
            password=url.password,
            host=url.host,
            port=url.port,
            path=url.path[1:],
        )
    elif url.scheme in mysql_allowed_schemes:
        dsn = MySQLDsn.build(
            scheme='mysql+asyncmy',
            username=url.username,
            password=url.password,
            host=url.host,
            port=url.port,
            path=url.path[1:],
        )
    else:
        raise ServerError(msg='SQLDsn not allowed_schemes')

    return str(dsn)


def create_engine_and_session(url: str):
    try:
        # Database engine
        url = validate_database_dns(url)
        engine = create_async_engine(url, future=True, pool_pre_ping=True)
    except Exception as e:
        sys.exit()
    else:
        db_session = async_sessionmaker(bind=engine, autoflush=False, expire_on_commit=False)
        return engine, db_session



#!/usr/bin/env python3
from typing import Any, Dict, Generic, Type, TypeVar

from pydantic import BaseModel
from sqlalchemy import Join, Row, and_, delete, func, inspect, select, update
from sqlalchemy.ext.asyncio import AsyncSession

from ..models.helper import (
    _add_column_with_prefix,
    _auto_detect_join_condition,
    _extract_matching_columns_from_kwargs,
    _extract_matching_columns_from_schema,
)
from ..models.base import MappedBase

ModelType = TypeVar('ModelType', bound=MappedBase)
CreateSchemaType = TypeVar('CreateSchemaType', bound=BaseModel)
UpdateSchemaType = TypeVar('UpdateSchemaType', bound=BaseModel)
UpdateSchemaInternalType = TypeVar('UpdateSchemaInternalType', bound=BaseModel)
DeleteSchemaType = TypeVar('DeleteSchemaType', bound=BaseModel)


class CRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    """Base class for CRUD operations on a model.

    Parameters
    ----------
    model : Type[ModelType]
        The SQLAlchemy model type.
    """

    def __init__(self, model: Type[ModelType]) -> None:
        self.model = model

    async def get_with_condition(self, db: AsyncSession, condition):
        """
        Example:
            condition=[
                Jobs.job_confirmation_code == job_confirmation_code,
                Jobs.status = 1
            ]
        """
        query = select(self.model).filter(*condition)
        groups = await db.execute(query)
        return groups.scalars().first()

    async def get_(
        self,
        db: AsyncSession,
        *,
        pk: int | None = None,
    ) -> ModelType | None:
        """
        Get a piece of data by primary key id or name

        :param db:
        :param pk:
        :return:
        """
        result = await db.execute(select(self.model).where(self.model.id == pk))
        return result.scalars().first()

    async def exists_(self, db: AsyncSession, **kwargs: Any) -> bool:
        """Check if a record exists based on filters.

        Parameters
        ----------
        db : AsyncSession
            The SQLAlchemy async session.
        kwargs : dict
            Filters to apply to the query.

        Returns
        -------
        bool
            True if a record exists, False otherwise.
        """
        to_select = _extract_matching_columns_from_kwargs(model=self.model, kwargs=kwargs)
        stmt = select(*to_select).filter_by(**kwargs).limit(1)

        result = await db.execute(stmt)
        return result.first() is not None

    async def count_(self, db: AsyncSession, **kwargs: Any) -> int:
        """Count the records based on filters.

        Parameters
        ----------
        db : AsyncSession
            The SQLAlchemy async session.
        kwargs : dict
            Filters to apply to the query.

        Returns
        -------
        int
            Total count of records that match the applied filters.

        Note
        ----
        This method provides a quick way to get the count of records without retrieving the actual data.
        """
        if kwargs:
            conditions = [getattr(self.model, key) == value for key, value in kwargs.items()]
            combined_conditions = and_(*conditions)
            count_query = select(func.count()).select_from(self.model).filter(combined_conditions)
        else:
            count_query = select(func.count()).select_from(self.model)

        total_count: int = await db.scalar(count_query)

        return total_count

    async def get_multi_(
        self,
        db: AsyncSession,
        offset: int = 0,
        limit: int = 100,
        schema_to_select: type[BaseModel] | list[type[BaseModel]] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Fetch multiple records based on filters.

        Parameters
        ----------
        db : AsyncSession
            The SQLAlchemy async session.
        offset : int, optional
            Number of rows to skip before fetching. Default is 0.
        limit : int, optional
            Maximum number of rows to fetch. Default is 100.
        schema_to_select : Union[Type[BaseModel], list[Type[BaseModel]], None], optional
            Pydantic schema for selecting specific columns. Default is None to select all columns.
        kwargs : dict
            Filters to apply to the query.

        Returns
        -------
        dict[str, Any]
            Dictionary containing the fetched rows under 'data' key and total count under 'total_count'.
        """
        to_select = _extract_matching_columns_from_schema(model=self.model, schema=schema_to_select)
        stmt = select(*to_select).filter_by(**kwargs).offset(offset).limit(limit)

        result = await db.execute(stmt)
        data = [dict(row) for row in result.mappings()]

        total_count = await self.count_(db=db, **kwargs)

        return {'data': data, 'total_count': total_count}

    async def get_joined_(
        self,
        db: AsyncSession,
        join_model: type[ModelType],
        join_prefix: str | None = None,
        join_on: Join | None = None,
        schema_to_select: type[BaseModel] | list | None = None,
        join_schema_to_select: type[BaseModel] | list | None = None,
        join_type: str = 'left',
        **kwargs: Any,
    ) -> dict | None:
        """Fetches a single record with a join on another model. If 'join_on' is not provided, the method attempts
        to automatically detect the join condition using foreign key relationships.

        Parameters
        ----------
        db : AsyncSession
            The SQLAlchemy async session.
        join_model : Type[ModelType]
            The model to join with.
        join_prefix : Optional[str]
            Optional prefix to be added to all columns of the joined model. If None, no prefix is added.
        join_on : Join, optional
            SQLAlchemy Join object for specifying the ON clause of the join. If None, the join condition is
            auto-detected based on foreign keys.
        schema_to_select : Union[Type[BaseModel], list, None], optional
            Pydantic schema for selecting specific columns from the primary model.
        join_schema_to_select : Union[Type[BaseModel], list, None], optional
            Pydantic schema for selecting specific columns from the joined model.
        join_type : str, default "left"
            Specifies the type of join operation to perform. Can be "left" for a left outer join
            or "inner" for an inner join.
        kwargs : dict
            Filters to apply to the query.

        Returns
        -------
        dict | None
            The fetched database row or None if not found.

        Examples
        --------
        Simple example: Joining User and Tier models without explicitly providing join_on
        ```python
        result = await crud_user.get_joined(
            db=session, join_model=Tier, schema_to_select=UserSchema, join_schema_to_select=TierSchema
        )
        ```

        Complex example: Joining with a custom join condition, additional filter parameters, and a prefix
        ```python
        from sqlalchemy import and_

        result = await crud_user.get_joined(
            db=session,
            join_model=Tier,
            join_prefix="tier_",
            join_on=and_(User.tier_id == Tier.id, User.is_superuser == True),
            schema_to_select=UserSchema,
            join_schema_to_select=TierSchema,
            username="john_doe",
        )
        ```

        Return example: prefix added, no schema_to_select or join_schema_to_select
        ```python
        {
            "id": 1,
            "name": "John Doe",
            "username": "john_doe",
            "email": "johndoe@example.com",
            "hashed_password": "hashed_password_example",
            "profile_image_url": "https://profileimageurl.com/default.jpg",
            "uuid": "123e4567-e89b-12d3-a456-426614174000",
            "created_at": "2023-01-01T12:00:00",
            "updated_at": "2023-01-02T12:00:00",
            "deleted_at": null,
            "is_deleted": false,
            "is_superuser": false,
            "tier_id": 2,
            "tier_name": "Premium",
            "tier_created_at": "2022-12-01T10:00:00",
            "tier_updated_at": "2023-01-01T11:00:00",
        }
        ```
        """
        if join_on is None:
            join_on = _auto_detect_join_condition(self.model, join_model)

        primary_select = _extract_matching_columns_from_schema(model=self.model, schema=schema_to_select)
        join_select = []

        if join_schema_to_select:
            columns = _extract_matching_columns_from_schema(model=join_model, schema=join_schema_to_select)
        else:
            columns = inspect(join_model).c

        for column in columns:
            labeled_column = _add_column_with_prefix(column, join_prefix)
            if f'{join_prefix}{column.name}' not in [col.name for col in primary_select]:
                join_select.append(labeled_column)

        if join_type == 'left':
            stmt = select(*primary_select, *join_select).outerjoin(join_model, join_on)
        elif join_type == 'inner':
            stmt = select(*primary_select, *join_select).join(join_model, join_on)
        else:
            raise ValueError(f"Invalid join type: {join_type}. Only 'left' or 'inner' are valid.")

        for key, value in kwargs.items():
            if hasattr(self.model, key):
                stmt = stmt.where(getattr(self.model, key) == value)

        db_row = await db.execute(stmt)
        result: Row = db_row.first()
        if result:
            out: dict = dict(result._mapping)
            return out

        return None

    async def get_multi_joined_(
        self,
        db: AsyncSession,
        join_model: type[ModelType],
        join_prefix: str | None = None,
        join_on: Join | None = None,
        schema_to_select: type[BaseModel] | list[type[BaseModel]] | None = None,
        join_schema_to_select: type[BaseModel] | list[type[BaseModel]] | None = None,
        join_type: str = 'left',
        offset: int = 0,
        limit: int = 100,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Fetch multiple records with a join on another model, allowing for pagination.

        Parameters
        ----------
        db : AsyncSession
            The SQLAlchemy async session.
        join_model : Type[ModelType]
            The model to join with.
        join_prefix : Optional[str]
            Optional prefix to be added to all columns of the joined model. If None, no prefix is added.
        join_on : Join, optional
            SQLAlchemy Join object for specifying the ON clause of the join. If None, the join condition is
            auto-detected based on foreign keys.
        schema_to_select : Union[Type[BaseModel], list[Type[BaseModel]], None], optional
            Pydantic schema for selecting specific columns from the primary model.
        join_schema_to_select : Union[Type[BaseModel], list[Type[BaseModel]], None], optional
            Pydantic schema for selecting specific columns from the joined model.
        join_type : str, default "left"
            Specifies the type of join operation to perform. Can be "left" for a left outer join
            or "inner" for an inner join.
        offset : int, default 0
            The offset (number of records to skip) for pagination.
        limit : int, default 100
            The limit (maximum number of records to return) for pagination.
        kwargs : dict
            Filters to apply to the primary query.

        Returns
        -------
        dict[str, Any]
            A dictionary containing the fetched rows under 'data' key and total count under 'total_count'.

        Examples
        --------
        # Fetching multiple User records joined with Tier records, using left join
        users = await crud_user.get_multi_joined(
            db=session,
            join_model=Tier,
            join_prefix="tier_",
            schema_to_select=UserSchema,
            join_schema_to_select=TierSchema,
            offset=0,
            limit=10
        )
        """
        if join_on is None:
            join_on = _auto_detect_join_condition(self.model, join_model)

        primary_select = _extract_matching_columns_from_schema(model=self.model, schema=schema_to_select)
        join_select = []

        if join_schema_to_select:
            columns = _extract_matching_columns_from_schema(model=join_model, schema=join_schema_to_select)
        else:
            columns = inspect(join_model).c

        for column in columns:
            labeled_column = _add_column_with_prefix(column, join_prefix)
            if f'{join_prefix}{column.name}' not in [col.name for col in primary_select]:
                join_select.append(labeled_column)

        if join_type == 'left':
            stmt = select(*primary_select, *join_select).outerjoin(join_model, join_on)
        elif join_type == 'inner':
            stmt = select(*primary_select, *join_select).join(join_model, join_on)
        else:
            raise ValueError(f"Invalid join type: {join_type}. Only 'left' or 'inner' are valid.")

        for key, value in kwargs.items():
            if hasattr(self.model, key):
                stmt = stmt.where(getattr(self.model, key) == value)

        stmt = stmt.offset(offset).limit(limit)

        db_rows = await db.execute(stmt)
        data = [dict(row._mapping) for row in db_rows]

        total_count = await self.count_(db=db, **kwargs)

        return {'data': data, 'total_count': total_count}

    async def create_(self, db: AsyncSession, obj_in: CreateSchemaType, user_id: int | None = None) -> None:
        """Create a new record in the database.

        :param db: AsyncSession
        :param obj_in: Pydantic Model class
        :param user_id:
        :return:
        """
        if user_id:
            create_data = self.model(**obj_in.model_dump(), create_user=user_id)
        else:
            create_data = self.model(**obj_in.model_dump())
        db.add(create_data)

    async def update_(
        self, db: AsyncSession, pk: int, obj_in: UpdateSchemaType | Dict[str, Any], user_id: int | None = None
    ) -> int:
        """
        Update a piece of data by primary key id

        :param db:
        :param pk:
        :param obj_in: Pydantic Model class or dictionary corresponding to database fields
        :param user_id:
        :return:
        """
        if isinstance(obj_in, dict):
            update_data = obj_in
        else:
            update_data = obj_in.model_dump(exclude_unset=True)
        if user_id:
            update_data.update({'update_user': user_id})
        result = await db.execute(update(self.model).where(self.model.id == pk).values(**update_data))
        return result.rowcount

    async def delete_(self, db: AsyncSession, pk: int, *, del_flag: int | None = None) -> int:
        """
        Delete a piece of data by primary key id

        :param db:
        :param pk:
        :param del_flag:
        :return:
        """
        if del_flag is None:
            result = await db.execute(delete(self.model).where(self.model.id == pk))
        else:
            assert del_flag == 1, 'Delete error, del_flag parameter can only be 1'
            result = await db.execute(update(self.model).where(self.model.id == pk).values(del_flag=del_flag))
        return result.rowcount



"""Facebook SSO Login Helper."""

from typing import TYPE_CHECKING, ClassVar, Optional

from .base import DiscoveryDocument, OpenID, SSOBase
from .constant import ProviderEnum

if TYPE_CHECKING:
    import httpx  # pragma: no cover


class FacebookSSO(SSOBase):
    """Class providing login via Facebook OAuth."""

    provider = ProviderEnum.FACEBOOK
    base_url = 'https://graph.facebook.com/v19.0'
    scope: ClassVar = ['email']

    async def get_discovery_document(self) -> DiscoveryDocument:
        """Get document containing handy urls."""
        return {
            'authorization_endpoint': 'https://www.facebook.com/v9.0/dialog/oauth',
            'token_endpoint': f'{self.base_url}/oauth/access_token',
            'userinfo_endpoint': f'{self.base_url}/me?fields=id,name,email,first_name,last_name,picture',
        }

    async def openid_from_response(self, response: dict, session: Optional['httpx.AsyncClient'] = None) -> OpenID:
        """Return OpenID from user information provided by Facebook."""

        return OpenID(
            email=response.get('email'),
            first_name=response.get('first_name'),
            last_name=response.get('last_name'),
            display_name=response.get('name'),
            provider=self.provider,
            id=response.get('id'),
            picture=response.get('picture', {}).get('data', {}).get('url', None),
        )



"""Gitlab SSO Oauth Helper class."""

from typing import TYPE_CHECKING, ClassVar, List, Optional, Tuple, Union
from urllib.parse import urljoin

import pydantic

from .base import DiscoveryDocument, OpenID, SSOBase
from .constant import ProviderEnum

if TYPE_CHECKING:
    import httpx  # pragma: no cover


class GitlabSSO(SSOBase):
    """Class providing login via Gitlab SSO."""

    provider = ProviderEnum.GITLAB
    scope: ClassVar = ['read_user', 'openid', 'profile']
    additional_headers: ClassVar = {'accept': 'application/json'}
    base_endpoint_url = 'https://gitlab.com'

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uri: Optional[Union[pydantic.AnyHttpUrl, str]] = None,
        allow_insecure_http: bool = False,
        use_state: bool = False,  # TODO: Remove use_state argument
        scope: Optional[List[str]] = None,
        base_endpoint_url: Optional[str] = None,
    ) -> None:
        super().__init__(
            client_id,
            client_secret,
            redirect_uri,
            allow_insecure_http,
            use_state,  # TODO: Remove use_state argument
            scope,
        )
        self.base_endpoint_url = base_endpoint_url or self.base_endpoint_url

    async def get_discovery_document(self) -> DiscoveryDocument:
        """Override the discovery document method to return Yandex OAuth endpoints."""
        return {
            'authorization_endpoint': urljoin(self.base_endpoint_url, '/oauth/authorize'),
            'token_endpoint': urljoin(self.base_endpoint_url, '/oauth/token'),
            'userinfo_endpoint': urljoin(self.base_endpoint_url, '/api/v4/user'),
        }

    def _parse_name(self, full_name: Optional[str]) -> Tuple[Union[str, None], Union[str, None]]:
        """Parses the full name from Gitlab into the first and last name."""
        if not full_name or not isinstance(full_name, str):
            return None, None

        name_parts = full_name.split()

        if len(name_parts) == 1:
            return name_parts[0], None

        first_name = name_parts[0]
        last_name = ' '.join(name_parts[1:])
        return first_name, last_name

    async def openid_from_response(self, response: dict, session: Optional['httpx.AsyncClient'] = None) -> OpenID:
        """Converts Gitlab user info response to OpenID object."""
        first_name, last_name = self._parse_name(response.get('name'))

        return OpenID(
            email=response['email'],
            provider=self.provider,
            id=str(response['id']),
            first_name=first_name,
            last_name=last_name,
            display_name=response['username'],
            picture=response['avatar_url'],
        )



#!/usr/bin/env python3
"""
Global business exception class

When business code executes abnormally, you can use raise xxxError to trigger internal errors.
It implements exceptions with background tasks as much as possible, but it does not apply to **custom response status codes**
If you are required to use **custom response status code**, you can return directly by return await response_base.fail(res=CustomResponseCode.xxx)
"""  # noqa: E501

from typing import Any

from fastapi import HTTPException

from .error_codes import StandardResponseCode


class BaseExceptionMixin(Exception):
    code: int

    def __init__(self, *, msg: str = None, data: Any = None):
        self.msg = msg
        self.data = data


class HTTPError(HTTPException):
    def __init__(self, *, code: int, msg: Any = None, headers: dict[str, Any] | None = None):
        super().__init__(status_code=code, detail=msg, headers=headers)


class RequestError(BaseExceptionMixin):
    code = StandardResponseCode.HTTP_400

    def __init__(self, *, msg: str = 'Bad Request', data: Any = None):
        super().__init__(msg=msg, data=data)


class ForbiddenError(BaseExceptionMixin):
    code = StandardResponseCode.HTTP_403

    def __init__(self, *, msg: str = 'Forbidden', data: Any = None):
        super().__init__(msg=msg, data=data)


class NotFoundError(BaseExceptionMixin):
    code = StandardResponseCode.HTTP_404

    def __init__(self, *, msg: str = 'Not Found', data: Any = None):
        super().__init__(msg=msg, data=data)


class UnprocessedContentError(BaseExceptionMixin):
    code = StandardResponseCode.HTTP_422

    def __init__(self, *, msg: str = 'Unprocessed Content', data: Any = None):
        super().__init__(msg=msg, data=data)


class ServerError(BaseExceptionMixin):
    code = StandardResponseCode.HTTP_500

    def __init__(self, *, msg: str = 'Internal Server Error', data: Any = None):
        super().__init__(msg=msg, data=data)


class GatewayError(BaseExceptionMixin):
    code = StandardResponseCode.HTTP_502

    def __init__(self, *, msg: str = 'Bad Gateway', data: Any = None):
        super().__init__(msg=msg, data=data)


class AuthorizationError(BaseExceptionMixin):
    code = StandardResponseCode.HTTP_401

    def __init__(self, *, msg: str = 'Permission Denied', data: Any = None):
        super().__init__(msg=msg, data=data)


class TokenError(HTTPError):
    code = StandardResponseCode.HTTP_401

    def __init__(self, *, msg: str = 'Not Authenticated', headers: dict[str, Any] | None = None):
        super().__init__(code=self.code, msg=msg, headers=headers or {'WWW-Authenticate': 'Bearer'})



#!/usr/bin/env python3

from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from .helper import generate_uuid
from ..models.base import Base, id_key


class SocialAccount(Base):
    """User table"""

    __tablename__ = 't_social_accounts'

    id: Mapped[id_key] = mapped_column(init=False)
    uuid: Mapped[str] = mapped_column(String(50), init=False, default_factory=generate_uuid, unique=True)
    login_id: Mapped[str] = mapped_column(String(255), unique=True, index=True, comment='Username')
    password: Mapped[str] = mapped_column(String(255), comment='Password')
    provider: Mapped[str] = mapped_column(String(255), comment='Provider')
    user_id: Mapped[str] = mapped_column(String(255), nullable=True)



#!/usr/bin/env python3

from fastapi import APIRouter

from ..schemas.auth import LoginParam
from ..schemas.response_schema import response_base
from ..schemas.token import GetLoginToken
from ..services.auth import AuthService

auth_router = APIRouter(prefix='/auth', tags=['Auth'])


@auth_router.post(path='/login', summary='auth')
async def user_auth(login_request: LoginParam):
    (access_token, refresh_token, access_expire, refresh_expire) = await AuthService().login(login_request)

    data = GetLoginToken(
        access_token=access_token,
        refresh_token=refresh_token,
        access_token_expire_time=access_expire,
        refresh_token_expire_time=refresh_expire,
    )
    return await response_base.success(data=data)



"""Helper functions to generate state param."""

import base64
import os


def generate_random_state(length: int = 64) -> str:
    """Generate a url-safe string to use as a state."""
    bytes_length = int(length * 3 / 4)
    return base64.urlsafe_b64encode(os.urandom(bytes_length)).decode('utf-8')



from typing import Annotated

from sqlalchemy.dialects.postgresql import TIMESTAMP
from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, declared_attr, mapped_column

from ..conf import settings
from ..oauth.timezone import DatetimeUTC, timezone

# Universal Mapped type primary key, needs to be added manually, refer to the following usage methods
# MappedBase -> id: Mapped[id_key]
# DataClassBase && Base -> id: Mapped[id_key] = mapped_column(init=False)
id_key = Annotated[
    int, mapped_column(primary_key=True, index=True, autoincrement=True, sort_order=-999, comment='Primary key id')
]


class DateTimeMixin(MappedAsDataclass):
    """Date and time Mixin data class"""

    created_time: Mapped[DatetimeUTC] = mapped_column(
        type_=TIMESTAMP(timezone=True),
        init=False,
        default_factory=timezone.now,
        sort_order=999,
        comment='Creation time',
    )
    updated_time: Mapped[DatetimeUTC | None] = mapped_column(
        type_=TIMESTAMP(timezone=True), init=False, onupdate=timezone.now, sort_order=999, comment='Update time'
    )


class MappedBase(DeclarativeBase):
    """
    Declarative base class, the original DeclarativeBase class,
    exists as the parent class of all base or data model classes

    `DeclarativeBase <https://docs.sqlalchemy.org/en/20/orm/declarative_config.html>`__
    `mapped_column() <https://docs.sqlalchemy.org/en/20/orm/mapping_api.html#sqlalchemy.orm.mapped_column>`__
    """

    @declared_attr.directive
    def __tablename__(cls) -> str:
        return cls.__name__.lower()


class DataClassBase(MappedAsDataclass, MappedBase):
    """
    Declarative data class base class, which will come with data class integration,
    allowing more advanced configuration, but you must pay attention to some of its characteristics,
    especially when used with DeclarativeBase

    `MappedAsDataclass <https://docs.sqlalchemy.org/en/20/orm/dataclasses.html#orm-declarative-native-dataclasses>`__
    """

    __abstract__ = True


class Base(DataClassBase, DateTimeMixin):
    """
    Declarative Mixin data class base class, with data class integration, and contains
    MiXin data class basic table structure, you can simply understand it as a data class base class
    containing basic table structure
    """

    __abstract__ = True


async def create_table():
    """Create database table"""
    async with settings.ENGINE.begin() as coon:
        await coon.run_sync(MappedBase.metadata.create_all)



from ..oauth import jwt
from ..base import errors
from .. import settings
from ..cruds.crud_social_account import SocialAccountDao
from ..schemas.auth import LoginParam


class AuthService:
    def __init__(self, provider='password'):
        self.provider = provider

    async def login(self, request: LoginParam):
        async with settings.DB_SESSION() as db:
            account = await SocialAccountDao.get_by_login_id_and_provider(
                db=db, login_id=request.username, provider=self.provider
            )

            await self.validate_login(account, request)
            access_token, access_token_expire_time = await jwt.create_access_token(
                str(account.user_id), provider=self.provider
            )
            refresh_token, refresh_token_expire_time = await jwt.create_refresh_token(
                str(account.user_id), access_token_expire_time, provider=self.provider
            )
            return (access_token, refresh_token, access_token_expire_time, refresh_token_expire_time)

    @staticmethod
    async def validate_login(current_account, data):
        if not current_account:
            raise errors.NotFoundError(msg='Please enter a valid username or password')
        else:
            if data.password and not await jwt.password_verify(data.password, current_account.password):
                raise errors.AuthorizationError(msg='Please enter a valid username or password')



#!/usr/bin/env python3
from datetime import datetime, timedelta
from typing import Tuple

from asgiref.sync import sync_to_async
from fastapi import Depends, Request
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
from fastapi.security.utils import get_authorization_scheme_param
from jose import jwt as jose_jwt
from passlib.context import CryptContext

from .errors import InvalidTokenError, TokenExpiredError
from .timezone import DatetimeUTC, timezone
from ..conf import settings


class JWTBearer(HTTPBearer):
    def __init__(self, auto_error: bool = True):
        super(JWTBearer, self).__init__(auto_error=auto_error)

    async def __call__(self, request: Request):
        credentials: HTTPAuthorizationCredentials = await super(JWTBearer, self).__call__(request)
        if credentials:
            if not credentials.scheme == 'Bearer':
                raise InvalidTokenError()

            if not await self.verify_jwt(credentials.credentials):
                raise InvalidTokenError()

            return credentials.credentials
        else:
            raise InvalidTokenError()

    async def verify_jwt(self, jwtoken: str) -> bool:
        isTokenValid: bool = False

        try:
            payload = await jwt_decode(jwtoken)
        except:
            payload = None

        if payload:
            isTokenValid = True

        return isTokenValid


pwd_context = CryptContext(schemes=['bcrypt'], deprecated='auto')

oauth2_schema = JWTBearer()


@sync_to_async
def get_hash_password(password: str) -> str:
    """
    Encrypt passwords using the hash algorithm

    :param password:
    :return:
    """
    return pwd_context.hash(password)


@sync_to_async
def password_verify(plain_password: str, hashed_password: str) -> bool:
    """
    Password verification

    :param plain_password: The password to verify
    :param hashed_password: The hash ciphers to compare
    :return:
    """
    return pwd_context.verify(plain_password, hashed_password)


async def create_access_token(sub: str, expires_delta: timedelta | None = None, **kwargs) -> tuple[str, DatetimeUTC]:
    """
    Generate encryption token

    :param sub: The subject/userid of the JWT
    :param expires_delta: Increased expiry time
    :return:
    """
    if expires_delta:
        expire = timezone.now() + expires_delta
    else:
        expire = timezone.now() + timedelta(seconds=settings.TOKEN_EXPIRE_SECONDS)

    multi_login = kwargs.pop('multi_login', None)
    to_encode = {'exp': expire, 'sub': sub, **kwargs}
    token = jose_jwt.encode(to_encode, settings.TOKEN_SECRET_KEY, settings.TOKEN_ALGORITHM)

    return token, expire


async def create_refresh_token(sub: str, expire_time: DatetimeUTC | None = None, **kwargs) -> tuple[str, DatetimeUTC]:
    """
    Generate encryption refresh token, only used to create a new token

    :param sub: The subject/userid of the JWT
    :param expire_time: expiry time
    :return:
    """
    if expire_time:
        expire = expire_time + timedelta(seconds=settings.TOKEN_REFRESH_EXPIRE_SECONDS)
        expire_datetime = timezone.f_datetime(expire)
        current_datetime = timezone.now()

        if expire_datetime < current_datetime:
            raise TokenExpiredError()

    else:
        expire = timezone.now() + timedelta(seconds=settings.TOKEN_REFRESH_EXPIRE_SECONDS)

    multi_login = kwargs.pop('multi_login', None)
    to_encode = {'exp': expire, 'sub': sub, **kwargs}
    refresh_token = jose_jwt.encode(to_encode, settings.TOKEN_SECRET_KEY, settings.TOKEN_ALGORITHM)

    return refresh_token, expire


async def create_new_token(sub: str, token: str, refresh_token: str, **kwargs) -> tuple[str, str, datetime, datetime]:
    """
    Generate new token

    :param sub:
    :param token
    :param refresh_token:
    :return:
    """

    new_access_token, new_access_token_expire_time = await create_access_token(sub, **kwargs)
    new_refresh_token, new_refresh_token_expire_time = await create_refresh_token(sub, **kwargs)

    return new_access_token, new_refresh_token, new_access_token_expire_time, new_refresh_token_expire_time


@sync_to_async
def get_token(request: Request) -> str:
    """
    Get token for request header

    :return:
    """
    authorization = request.headers.get('Authorization')
    scheme, token = get_authorization_scheme_param(authorization)
    if not authorization or scheme.lower() != 'bearer':
        raise InvalidTokenError()

    return token


@sync_to_async
def jwt_decode(token: str) -> Tuple:
    """
    Decode token

    :param token:
    :return:
    """
    try:
        payload = jose_jwt.decode(token, settings.TOKEN_SECRET_KEY, algorithms=[settings.TOKEN_ALGORITHM])
        user_id = int(payload.get('sub'))
        if not user_id:
            raise InvalidTokenError()

        provider = payload.get('provider')
    except (jose_jwt.ExpiredSignatureError, jose_jwt.JWTError, Exception):
        raise InvalidTokenError()

    return user_id, provider


async def jwt_authentication(token: str) -> dict[str, int]:
    """
    JWT authentication

    :param token:
    :param jwt_type:
    :return:
    """
    user_id, provider = await jwt_decode(token)
    return {'sub': user_id, 'provider': provider}


async def jwt_encryption_token(
    payload: dict,
    expired_time: int | None = None,  # seconds
) -> tuple[str, datetime]:
    """
    Generate encryption token
    """
    if expired_time:
        expire = timezone.now() + timedelta(seconds=expired_time)
    else:
        expire = timezone.now() + timedelta(seconds=settings.TOKEN_EXPIRE_SECONDS)

    token = jose_jwt.encode(payload, settings.TOKEN_SECRET_KEY, settings.TOKEN_ALGORITHM)

    return token, expire


async def jwt_decryption_token(token: str) -> dict:
    """
    Decode token
    """
    try:
        payload = jose_jwt.decode(token, settings.TOKEN_SECRET_KEY, algorithms=[settings.TOKEN_ALGORITHM])
    except (jose_jwt.ExpiredSignatureError, jose_jwt.JWTError, Exception):
        raise InvalidTokenError()

    return payload


DependsJwtAuth = Depends(oauth2_schema)



"""Google SSO Login Helper."""

from typing import ClassVar, Optional

import httpx

from .base import DiscoveryDocument, OpenID, SSOBase, SSOLoginError
from .constant import ProviderEnum


class GoogleSSO(SSOBase):
    """Class providing login via Google OAuth."""

    discovery_url = 'https://accounts.google.com/.well-known/openid-configuration'
    provider = ProviderEnum.GOOGLE
    scope: ClassVar = ['openid', 'email', 'profile']

    async def openid_from_response(self, response: dict, session: Optional['httpx.AsyncClient'] = None) -> OpenID:
        """Return OpenID from user information provided by Google."""
        if response.get('email_verified'):
            return OpenID(
                email=response.get('email'),
                provider=self.provider,
                id=response.get('sub'),
                first_name=response.get('given_name'),
                last_name=response.get('family_name'),
                display_name=response.get('name'),
                picture=response.get('picture'),
            )
        raise SSOLoginError(401, f"User {response.get('email')} is not verified with Google")

    async def get_discovery_document(self) -> DiscoveryDocument:
        """Get document containing handy urls."""
        async with httpx.AsyncClient() as session:
            response = await session.get(self.discovery_url)
            content = response.json()
            return content



class StandardResponseCode:
    """Standard response status code"""

    """
    HTTP codes
    See HTTP Status Code Registry:
    https://www.iana.org/assignments/http-status-codes/http-status-codes.xhtml

    And RFC 2324 - https://tools.ietf.org/html/rfc2324
    """
    HTTP_100 = 100  # CONTINUE
    HTTP_101 = 101  # SWITCHING_PROTOCOLS
    HTTP_102 = 102  # PROCESSING
    HTTP_103 = 103  # EARLY_HINTS
    HTTP_200 = 200  # OK
    HTTP_201 = 201  # CREATED
    HTTP_202 = 202  # ACCEPTED
    HTTP_203 = 203  # NON_AUTHORITATIVE_INFORMATION
    HTTP_204 = 204  # NO_CONTENT
    HTTP_205 = 205  # RESET_CONTENT
    HTTP_206 = 206  # PARTIAL_CONTENT
    HTTP_207 = 207  # MULTI_STATUS
    HTTP_208 = 208  # ALREADY_REPORTED
    HTTP_226 = 226  # IM_USED
    HTTP_300 = 300  # MULTIPLE_CHOICES
    HTTP_301 = 301  # MOVED_PERMANENTLY
    HTTP_302 = 302  # FOUND
    HTTP_303 = 303  # SEE_OTHER
    HTTP_304 = 304  # NOT_MODIFIED
    HTTP_305 = 305  # USE_PROXY
    HTTP_307 = 307  # TEMPORARY_REDIRECT
    HTTP_308 = 308  # PERMANENT_REDIRECT
    HTTP_400 = 400  # BAD_REQUEST
    HTTP_401 = 401  # UNAUTHORIZED
    HTTP_402 = 402  # PAYMENT_REQUIRED
    HTTP_403 = 403  # FORBIDDEN
    HTTP_404 = 404  # NOT_FOUND
    HTTP_405 = 405  # METHOD_NOT_ALLOWED
    HTTP_406 = 406  # NOT_ACCEPTABLE
    HTTP_407 = 407  # PROXY_AUTHENTICATION_REQUIRED
    HTTP_408 = 408  # REQUEST_TIMEOUT
    HTTP_409 = 409  # CONFLICT
    HTTP_410 = 410  # GONE
    HTTP_411 = 411  # LENGTH_REQUIRED
    HTTP_412 = 412  # PRECONDITION_FAILED
    HTTP_413 = 413  # REQUEST_ENTITY_TOO_LARGE
    HTTP_414 = 414  # REQUEST_URI_TOO_LONG
    HTTP_415 = 415  # UNSUPPORTED_MEDIA_TYPE
    HTTP_416 = 416  # REQUESTED_RANGE_NOT_SATISFIABLE
    HTTP_417 = 417  # EXPECTATION_FAILED
    HTTP_418 = 418  # UNUSED
    HTTP_421 = 421  # MISDIRECTED_REQUEST
    HTTP_422 = 422  # UNPROCESSABLE_CONTENT
    HTTP_423 = 423  # LOCKED
    HTTP_424 = 424  # FAILED_DEPENDENCY
    HTTP_425 = 425  # TOO_EARLY
    HTTP_426 = 426  # UPGRADE_REQUIRED
    HTTP_427 = 427  # UNASSIGNED
    HTTP_428 = 428  # PRECONDITION_REQUIRED
    HTTP_429 = 429  # TOO_MANY_REQUESTS
    HTTP_430 = 430  # Unassigned
    HTTP_431 = 431  # REQUEST_HEADER_FIELDS_TOO_LARGE
    HTTP_451 = 451  # UNAVAILABLE_FOR_LEGAL_REASONS
    HTTP_500 = 500  # INTERNAL_SERVER_ERROR
    HTTP_501 = 501  # NOT_IMPLEMENTED
    HTTP_502 = 502  # BAD_GATEWAY
    HTTP_503 = 503  # SERVICE_UNAVAILABLE
    HTTP_504 = 504  # GATEWAY_TIMEOUT
    HTTP_505 = 505  # HTTP_VERSION_NOT_SUPPORTED
    HTTP_506 = 506  # VARIANT_ALSO_NEGOTIATES
    HTTP_507 = 507  # INSUFFICIENT_STORAGE
    HTTP_508 = 508  # LOOP_DETECTED
    HTTP_509 = 509  # UNASSIGNED
    HTTP_510 = 510  # NOT_EXTENDED
    HTTP_511 = 511  # NETWORK_AUTHENTICATION_REQUIRED



"""Line SSO Login Helper."""

from typing import TYPE_CHECKING, ClassVar, Optional

from .base import DiscoveryDocument, OpenID, SSOBase
from .constant import ProviderEnum

if TYPE_CHECKING:
    import httpx  # pragma: no cover


class LineSSO(SSOBase):
    """Class providing login via Line OAuth."""

    provider = ProviderEnum.LINE
    base_url = 'https://api.line.me/oauth2/v2.1'
    scope: ClassVar = ['email', 'profile', 'openid']

    async def get_discovery_document(self) -> DiscoveryDocument:
        """Get document containing handy urls."""
        return {
            'authorization_endpoint': 'https://access.line.me/oauth2/v2.1/authorize',
            'token_endpoint': f'{self.base_url}/token',
            'userinfo_endpoint': f'{self.base_url}/userinfo',
        }

    async def openid_from_response(self, response: dict, session: Optional['httpx.AsyncClient'] = None) -> OpenID:
        """Return OpenID from user information provided by Line."""
        return OpenID(
            email=response.get('email'),
            first_name=None,
            last_name=None,
            display_name=response.get('name'),
            provider=self.provider,
            id=response.get('sub'),
            picture=response.get('picture'),
        )



#!/usr/bin/env python3

import secrets
from enum import Enum
from functools import lru_cache
from typing import Optional, Any

from pydantic_settings import BaseSettings


class LogLevel(str, Enum):
    """Possible log levels."""

    NOTSET = 'NOTSET'
    DEBUG = 'DEBUG'
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'
    FATAL = 'FATAL'


class SettingsSocial(BaseSettings):
    # Env Config
    GOOGLE_CLIENT_ID: Optional[str] = None
    GOOGLE_CLIENT_SECRET: Optional[str] = None

    GITHUB_CLIENT_ID: Optional[str] = None
    GITHUB_CLIENT_SECRET: Optional[str] = None

    FACEBOOK_CLIENT_ID: Optional[str] = None
    FACEBOOK_CLIENT_SECRET: Optional[str] = None

    FIBIT_CLIENT_ID: Optional[str] = None
    FIBIT_CLIENT_SECRET: Optional[str] = None

    GITLAB_CLIENT_ID: Optional[str] = None
    GITLAB_CLIENT_SECRET: Optional[str] = None
    GITLAB_BASE_ENDPOINT_URL: Optional[str] = None

    LINE_CLIENT_ID: Optional[str] = None
    LINE_CLIENT_SECRET: Optional[str] = None

    LINKEDIN_CLIENT_ID: Optional[str] = None
    LINKEDIN_CLIENT_SECRET: Optional[str] = None

    NOTION_CLIENT_ID: Optional[str] = None
    NOTION_CLIENT_SECRET: Optional[str] = None

    TWITTER_CLIENT_ID: Optional[str] = None
    TWITTER_CLIENT_SECRET: Optional[str] = None

    # DateTime
    DATETIME_TIMEZONE: str = 'UTC/GTM'
    DATETIME_FORMAT: str = '%Y-%m-%d %H:%M:%S'

    # Token
    TOKEN_SECRET_KEY: str = secrets.token_urlsafe(32)
    TOKEN_ALGORITHM: str = 'HS256'  # algorithm
    TOKEN_EXPIRE_SECONDS: int = 60 * 60 * 24  # Expiration time, unit: seconds 60 * 60 * 24 * 1
    TOKEN_REFRESH_EXPIRE_SECONDS: int = 60 * 60 * 24 * 7  # Refresh expiration time, unit: seconds

    # DB
    ENGINE: Optional[Any] = None
    DB_SESSION: Optional[Any] = None

    USER_MODEL: Optional[Any] = None


@lru_cache
def get_settings() -> SettingsSocial:
    """Read configuration optimization writing method"""
    return SettingsSocial()


settings = get_settings()



#!/usr/bin/env python3
from typing import Annotated

from fastapi import APIRouter, Request
from fastapi.params import Query, Path

from ..schemas.response_schema import response_base
from ..schemas.token import GetLoginToken
from ..services.social import SocialService

social_router = APIRouter(prefix='/social', tags=['Social'])


@social_router.get('/{provider}/auth-init')
async def social_auth_url_init(provider: Annotated[str, Path(...)], redirect_uri: Annotated[str, Query()]):
    """Initialize auth and redirect"""
    service = SocialService(provider)
    url = await service.build_auth_url(redirect_uri)
    return await response_base.success(data={'url': url})


@social_router.get('/{provider}/auth')
async def social_auth(
    provider: Annotated[str, Path(...)],
    request: Request,
    redirect_uri: Annotated[str, Query()],
    code: Annotated[str, Query()],
    state: Annotated[str, Query()],
):
    """Initialize auth and redirect"""
    service = SocialService(provider)
    (access_token, refresh_token, access_expire, refresh_expire) = await service.verify_login(request, redirect_uri)
    data = GetLoginToken(
        access_token=access_token,
        refresh_token=refresh_token,
        access_token_expire_time=access_expire,
        refresh_token_expire_time=refresh_expire,
    )
    return await response_base.success(data=data)



from typing import Any

from ...base.errors import BaseExceptionMixin


class SSOExceptionError(BaseExceptionMixin):
    code = 400

    def __init__(
        self,
        *,
        msg: str = 'SSO Exception Error',
        data: Any = None,
    ):
        super().__init__(msg=msg, data=data)



#!/usr/bin/env python3
from datetime import UTC, datetime
from datetime import timezone as tzz
from typing import Annotated, Any

from dateutil.parser import parse
from pydantic import AfterValidator, AwareDatetime, BeforeValidator, PlainSerializer, WithJsonSchema

from conf import settings


class TimeZone:
    def __init__(self, tz: str = None):
        self.tz_info = UTC

    def now(self) -> datetime:
        """
        Get time zone time

        :return:
        """
        return datetime.now(self.tz_info)

    def f_datetime(self, dt: datetime) -> datetime:
        """
        datetime Time to time zone time

        :param dt:
        :return:
        """
        return dt.astimezone(self.tz_info)

    def f_str(self, date_str: str, format_str: str = settings.DATETIME_FORMAT) -> datetime:
        """
        Convert time string to time zone

        :param date_str:
        :param format_str:
        :return:
        """
        return datetime.strptime(date_str, format_str).replace(tzinfo=self.tz_info)

    def f_timestamp(self, timestamp: int) -> datetime:
        """
        Convert timestamp to time zone

        :param timestamp:
        :return:
        """
        timestamp = int(timestamp) / 1000 if len(str(timestamp)) == 13 else int(timestamp)
        return datetime.fromtimestamp(timestamp)

    @staticmethod
    def f_datetime_t_timestamp(dt: datetime) -> int:
        """
        Convert datetime to timestamp

        :param dt:
        :return:
        """
        return int(dt.timestamp() * 1000)

    @staticmethod
    def fix_postgres_datetime(v: str) -> str:
        if isinstance(v, str):
            if '+' in v:
                dt, tz = v.split('+')
                if len(tz) == 2:
                    return f'{dt}+{tz}:00'
        return v

    def validate_timestamp(self, v: Any) -> int:
        """Make from naive datetime a timezone aware (with UTC timezone)."""
        if isinstance(v, (float, int)) or (isinstance(v, str) and v.isnumeric()):
            # parse value to datetime
            v = self.f_timestamp(int(v))
        if isinstance(v, str):
            try:
                v = self.f_str(self.fix_postgres_datetime(v), '%Y-%m-%d %H:%M:%S.%f%z')
            except ValueError:
                v = self.f_str(self.fix_postgres_datetime(v), '%Y-%m-%d %H:%M:%S%z')
        # if datetime is naive, just replace it to UTC
        if v.tzinfo is None:
            result = v.replace(tzinfo=UTC)
        # else convert to utc
        else:
            result = v.astimezone(UTC)

        return int(result.timestamp() * 1000)

    @staticmethod
    def validate_utc(dt: AwareDatetime) -> AwareDatetime:
        """Validate that the pydantic.AwareDatetime is in UTC."""
        if dt.tzinfo.utcoffset(dt) != tzz.utc.utcoffset(dt):
            raise ValueError('Timezone must be UTC')
        return dt

    def convert_datetime_string_to_timezone(self, datetime_str: str) -> datetime:
        return parse(datetime_str).astimezone(tz=self.tz_info)


timezone: TimeZone = TimeZone()

DatetimeUTC = Annotated[
    AwareDatetime, AfterValidator(timezone.validate_utc), BeforeValidator(timezone.fix_postgres_datetime)
]

Timestamp = Annotated[
    int,
    PlainSerializer(func=lambda val: val, return_type=int),
    BeforeValidator(func=lambda val: timezone.validate_timestamp(v=val)),
    WithJsonSchema(
        json_schema={'type': 'number', 'examples': [1704067200000.345, 1704067200000], 'example': 1704067200000.785},
        mode='serialization',
    ),
]



from sqlalchemy.ext.asyncio import AsyncSession

from ..conf import settings
from ..base import errors
from ..cruds.crud_social_account import SocialAccountDao
from ..schemas.social_account import RegisterSocialAccountParam, RegisterAccountParam


class UserService:
    def __init__(self):
        self.provider = 'password'

    async def register_user(self, obj: RegisterAccountParam, **kwargs):
        async with settings.DB_SESSION() as db:
            account = await SocialAccountDao.get_by_login_id_and_provider(
                db=db, login_id=obj.login_id, provider=self.provider
            )
            obj.provider = self.provider

            if account:
                raise errors.UnprocessedContentError(msg='Login ID is already registered.')

            account = await self.create_new_account(db, obj)
            await db.commit()

        return dict(user_id=account.user_id)

    async def create_new_account(self, db: AsyncSession, account_req: RegisterAccountParam):
        register_account = RegisterSocialAccountParam(
            login_id=account_req.login_id,
            user_id=account_req.user_id,
            provider=account_req.provider,
            password=account_req.password,
        )
        account = await SocialAccountDao.create(db, register_account)
        return account



from .google import GoogleService
from .gitlab import GitlabService
from .github import GithubService
from .facebook import FacebookService
from .line import LineService
from .fitbit import FitbitService
from .notion import NotionService
from .linkedin import LinkedInService
from .twitter import TwitterService



"""Github SSO Oauth Helper class."""

from typing import TYPE_CHECKING, ClassVar, Optional

from .base import DiscoveryDocument, OpenID, SSOBase
from .constant import ProviderEnum

if TYPE_CHECKING:
    import httpx  # pragma: no cover


class GithubSSO(SSOBase):
    """Class providing login via Github SSO."""

    provider = ProviderEnum.GITHUB
    scope: ClassVar = ['user:email']
    additional_headers: ClassVar = {'accept': 'application/json'}
    emails_endpoint = 'https://api.github.com/user/emails'

    async def get_discovery_document(self) -> DiscoveryDocument:
        return {
            'authorization_endpoint': 'https://github.com/login/oauth/authorize',
            'token_endpoint': 'https://github.com/login/oauth/access_token',
            'userinfo_endpoint': 'https://api.github.com/user',
        }

    async def _get_primary_email(self, session: Optional['httpx.AsyncClient'] = None) -> Optional[str]:
        """Attempt to get primary email from Github for a current user.
        The session received must be authenticated.
        """
        if not session:
            return None
        response = await session.get(self.emails_endpoint)
        if response.status_code != 200:
            return None
        emails = response.json()
        for email in emails:
            if email['primary']:
                return email['email']
        return None

    async def openid_from_response(self, response: dict, session: Optional['httpx.AsyncClient'] = None) -> OpenID:
        return OpenID(
            email=response.get('email') or (await self._get_primary_email(session)),
            provider=self.provider,
            id=str(response['id']),
            display_name=response['login'],
            picture=response['avatar_url'],
        )



#!/usr/bin/env python3
from typing import Optional

from fastapi import FastAPI
from loguru import logger

from .base.database import create_engine_and_session
from .conf import settings, SettingsSocial
from .models import SocialAccount
from .models.base import create_table
from .routers import router
from .social.sso.base import OpenID


async def register_auth_social_init(
    app: FastAPI,
    user_model,
    sqlalchemy_database_uri: str,
    token_secret_key: Optional[str] = None,
    token_algorithm: Optional[str] = None,
    token_expire_seconds: Optional[int] = None,
    token_refresh_expire_seconds: Optional[int] = None,
    google_client_id: Optional[str] = None,
    google_client_secret: Optional[str] = None,
    github_client_secret: Optional[str] = None,
    facebook_client_id: Optional[str] = None,
    facebook_client_secret: Optional[str] = None,
    fibit_client_id: Optional[str] = None,
    fibit_client_secret: Optional[str] = None,
    gitlab_client_id: Optional[str] = None,
    gitlab_client_secret: Optional[str] = None,
    gitlab_base_endpoint_url: Optional[str] = None,
    line_client_id: Optional[str] = None,
    line_client_secret: Optional[str] = None,
    linkedin_client_id: Optional[str] = None,
    linkedin_client_secret: Optional[str] = None,
    notion_client_id: Optional[str] = None,
    notion_client_secret: Optional[str] = None,
    twitter_client_id: Optional[str] = None,
    twitter_client_secret: Optional[str] = None,
):
    """
    Start initialization

    :return:
    """
    settings.USER_MODEL = user_model
    engine, db_session = create_engine_and_session(sqlalchemy_database_uri)
    settings.ENGINE = engine
    settings.DB_SESSION = db_session
    if token_secret_key:
        settings.TOKEN_SECRET_KEY = token_secret_key

    if token_algorithm:
        settings.TOKEN_ALGORITHM = token_algorithm

    if token_expire_seconds:
        settings.TOKEN_EXPIRE_SECONDS = token_expire_seconds

    if token_refresh_expire_seconds:
        settings.TOKEN_REFRESH_EXPIRE_SECONDS = token_refresh_expire_seconds

    settings.GOOGLE_CLIENT_ID = google_client_id
    settings.GOOGLE_CLIENT_SECRET = google_client_secret

    settings.GITHUB_CLIENT_ID = github_client_secret
    settings.GITHUB_CLIENT_SECRET = github_client_secret

    settings.FACEBOOK_CLIENT_ID = facebook_client_id
    settings.FACEBOOK_CLIENT_SECRET = facebook_client_secret

    settings.FIBIT_CLIENT_ID = fibit_client_id
    settings.FIBIT_CLIENT_SECRET = fibit_client_secret

    settings.GITLAB_CLIENT_ID = gitlab_client_id
    settings.GITLAB_CLIENT_SECRET = gitlab_client_secret
    settings.GITLAB_BASE_ENDPOINT_URL = gitlab_base_endpoint_url

    settings.LINE_CLIENT_ID = line_client_id
    settings.LINE_CLIENT_SECRET = line_client_secret

    settings.LINKEDIN_CLIENT_ID = linkedin_client_id
    settings.LINKEDIN_CLIENT_SECRET = linkedin_client_secret

    settings.NOTION_CLIENT_ID = notion_client_id
    settings.NOTION_CLIENT_SECRET = notion_client_secret

    settings.TWITTER_CLIENT_ID = twitter_client_id
    settings.TWITTER_CLIENT_SECRET = twitter_client_secret

    # Create database table
    logger.warning('Creating table first...')
    await create_table()
    logger.success('Create table successful')
    app.include_router(router)



"""LinkedIn SSO Oauth Helper class."""

from typing import TYPE_CHECKING, ClassVar, Dict, Optional

from .base import DiscoveryDocument, OpenID, SSOBase
from .constant import ProviderEnum

if TYPE_CHECKING:
    import httpx  # pragma: no cover


class LinkedInSSO(SSOBase):
    """Class providing login via LinkedIn SSO."""

    provider = ProviderEnum.LINKEDIN
    scope: ClassVar = ['openid', 'profile', 'email']
    additional_headers: ClassVar = {'accept': 'application/json'}

    @property
    def _extra_query_params(self) -> Dict:
        return {'client_secret': self.client_secret}

    async def get_discovery_document(self) -> DiscoveryDocument:
        return {
            'authorization_endpoint': 'https://www.linkedin.com/oauth/v2/authorization',
            'token_endpoint': 'https://www.linkedin.com/oauth/v2/accessToken',
            'userinfo_endpoint': 'https://api.linkedin.com/v2/userinfo',
        }

    async def openid_from_response(self, response: dict, session: Optional['httpx.AsyncClient'] = None) -> OpenID:
        return OpenID(
            email=response.get('email'),
            provider=self.provider,
            id=response.get('sub'),
            first_name=response.get('given_name'),
            last_name=response.get('family_name'),
            picture=response.get('picture'),
        )



"""SSO login base dependency."""

import json
import logging
import os
import warnings
from types import TracebackType
from typing import Any, ClassVar, Dict, List, Literal, Optional, Type, TypedDict, Union, overload
from urllib.parse import unquote

import httpx
import pydantic

from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import RedirectResponse

from ..sso.errors import SSOExceptionError
from ...oauth.web_application import WebApplicationClient
from ..sso.pkce import get_pkce_challenge_pair
from ..sso.state import generate_random_state

logger = logging.getLogger(__name__)


class DiscoveryDocument(TypedDict):
    """Discovery document."""

    authorization_endpoint: str
    token_endpoint: str
    userinfo_endpoint: str


class UnsetStateWarning(UserWarning):
    """Warning about unset state parameter."""


class ReusedOauthClientWarning(UserWarning):
    """Warning about reused oauth client instance."""


class SSOLoginError(HTTPException):
    """Raised when any login-related error ocurrs.

    Such as when user is not verified or if there was an attempt for fake login.
    """


class OpenID(pydantic.BaseModel):
    """Class (schema) to represent information got from sso provider in a common form."""

    id: Optional[str] = None
    email: Optional[pydantic.EmailStr] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    display_name: Optional[str] = None
    picture: Optional[str] = None
    provider: Optional[str] = None


class SSOBase:
    """Base class for all SSO providers."""

    provider: str = NotImplemented
    client_id: str = NotImplemented
    client_secret: str = NotImplemented
    redirect_uri: Optional[Union[pydantic.AnyHttpUrl, str]] = NotImplemented
    scope: ClassVar[List[str]] = []
    additional_headers: ClassVar[Optional[Dict[str, Any]]] = None
    uses_pkce: bool = False
    requires_state: bool = False

    _pkce_challenge_length: int = 96

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uri: Optional[Union[pydantic.AnyHttpUrl, str]] = None,
        allow_insecure_http: bool = False,
        use_state: bool = False,
        scope: Optional[List[str]] = None,
    ):
        """Base class (mixin) for all SSO providers."""
        self.client_id: str = client_id
        self.client_secret: str = client_secret
        self.redirect_uri: Optional[Union[pydantic.AnyHttpUrl, str]] = redirect_uri
        self.allow_insecure_http: bool = allow_insecure_http
        self._oauth_client: Optional[WebApplicationClient] = None
        self._generated_state: Optional[str] = None

        if self.allow_insecure_http:
            logger.debug('Initializing %s with allow_insecure_http=True', self.__class__.__name__)
            os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'

        # TODO: Remove use_state argument and attribute
        if use_state:
            warnings.warn(
                (
                    "Argument 'use_state' of SSOBase's constructor is deprecated and will be removed in "
                    "future releases. Use 'state' argument of individual methods instead."
                ),
                DeprecationWarning,
            )
            self.requires_state = True

        self._scope = scope or self.scope
        self._refresh_token: Optional[str] = None
        self._id_token: Optional[str] = None
        self._state: Optional[str] = None
        self._pkce_code_challenge: Optional[str] = None
        self._pkce_code_verifier: Optional[str] = None
        self._pkce_challenge_method = 'S256'

    @property
    def state(self) -> Optional[str]:
        """Retrieves the state as it was returned from the server.

        Warning:
            This will emit a warning if the state is unset, implying either that
            the server didn't return a state or `verify_and_process` hasn't been
            called yet.

        Returns:
            Optional[str]: The state parameter returned from the server.
        """
        if self._state is None:
            warnings.warn(
                "'state' parameter is unset. This means the server either "
                "didn't return state (was this expected?) or 'verify_and_process' hasn't been called yet.",
                UnsetStateWarning,
            )
        return self._state

    @property
    def oauth_client(self) -> WebApplicationClient:
        """Retrieves the OAuth Client to aid in generating requests and parsing responses.

        Raises:
            NotImplementedError: If the provider is not supported or `client_id` is not set.

        Returns:
            WebApplicationClient: OAuth client instance.
        """
        if self.client_id == NotImplemented:
            raise NotImplementedError(f'Provider {self.provider} not supported')  # pragma: no cover
        if self._oauth_client is None:
            self._oauth_client = WebApplicationClient(self.client_id)
        return self._oauth_client

    @property
    def access_token(self) -> Optional[str]:
        """Retrieves the access token from token endpoint.

        Returns:
            Optional[str]: The access token if available.
        """
        return self.oauth_client.access_token

    @property
    def refresh_token(self) -> Optional[str]:
        """Retrieves the refresh token if returned from provider.

        Returns:
            Optional[str]: The refresh token if available.
        """
        return self._refresh_token or self.oauth_client.refresh_token

    @property
    def id_token(self) -> Optional[str]:
        """Retrieves the id token if returned from provider.

        Returns:
            Optional[str]: The id token if available.
        """
        return self._id_token

    async def openid_from_response(self, response: dict, session: Optional[httpx.AsyncClient] = None) -> OpenID:
        """Converts a response from the provider's user info endpoint to an OpenID object.

        Args:
            response (dict): The response from the user info endpoint.
            session (Optional[httpx.AsyncClient]): The HTTPX AsyncClient session.

        Raises:
            NotImplementedError: If the provider is not supported.

        Returns:
            OpenID: The user information in a standardized format.
        """
        raise NotImplementedError(f'Provider {self.provider} not supported')

    async def get_discovery_document(self) -> DiscoveryDocument:
        """Retrieves the discovery document containing useful URLs.

        Raises:
            NotImplementedError: If the provider is not supported.

        Returns:
            DiscoveryDocument: A dictionary containing important endpoints like authorization, token and userinfo.
        """
        raise NotImplementedError(f'Provider {self.provider} not supported')

    @property
    async def authorization_endpoint(self) -> Optional[str]:
        """Return `authorization_endpoint` from discovery document."""
        discovery = await self.get_discovery_document()
        return discovery.get('authorization_endpoint')

    @property
    async def token_endpoint(self) -> Optional[str]:
        """Return `token_endpoint` from discovery document."""
        discovery = await self.get_discovery_document()
        return discovery.get('token_endpoint')

    @property
    async def userinfo_endpoint(self) -> Optional[str]:
        """Return `userinfo_endpoint` from discovery document."""
        discovery = await self.get_discovery_document()
        return discovery.get('userinfo_endpoint')

    async def get_login_url(
        self,
        *,
        redirect_uri: Optional[Union[pydantic.AnyHttpUrl, str]] = None,
        params: Optional[Dict[str, Any]] = None,
        state: Optional[str] = None,
    ) -> str:
        """Generates and returns the prepared login URL.

        Args:
            redirect_uri (Optional[str]): Overrides the `redirect_uri` specified on this instance.
            params (Optional[Dict[str, Any]]): Additional query parameters to add to the login request.
            state (Optional[str]): The state parameter for the OAuth 2.0 authorization request.

        Raises:
            ValueError: If `redirect_uri` is not provided either at construction or request time.

        Returns:
            str: The prepared login URL.
        """
        params = params or {}
        redirect_uri = redirect_uri or self.redirect_uri
        if redirect_uri is None:
            raise ValueError('redirect_uri must be provided, either at construction or request time')
        if self.uses_pkce and not all((self._pkce_code_verifier, self._pkce_code_challenge)):
            warnings.warn(
                f'{self.__class__.__name__!r} uses PKCE and no code was generated yet. '
                'Use SSO class as a context manager to get rid of this warning and possible errors.'
            )
        if self.requires_state and not state:
            if self._generated_state is None:
                warnings.warn(
                    f'{self.__class__.__name__!r} requires state in the request but none was provided nor '
                    'generated automatically. Use SSO as a context manager. The login process will most probably fail.'
                )
            state = self._generated_state
        request_uri = self.oauth_client.prepare_request_uri(
            await self.authorization_endpoint,
            redirect_uri=redirect_uri,
            state=state,
            scope=self._scope,
            code_challenge=self._pkce_code_challenge,
            code_challenge_method=self._pkce_challenge_method,
            **params,
        )
        return request_uri

    async def get_login_redirect(
        self,
        *,
        redirect_uri: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        state: Optional[str] = None,
    ) -> str:
        """Constructs and returns a redirect response to the login page of OAuth SSO provider.

        Args:
            redirect_uri (Optional[str]): Overrides the `redirect_uri` specified on this instance.
            params (Optional[Dict[str, Any]]): Additional query parameters to add to the login request.
            state (Optional[str]): The state parameter for the OAuth 2.0 authorization request.

        Returns:
            RedirectResponse: A URL response directing to the login page of the OAuth SSO provider.
        """
        if self.requires_state and not state:
            state = self._generated_state
        login_uri = await self.get_login_url(redirect_uri=redirect_uri, params=params, state=state)

        return login_uri

    @overload
    async def verify_and_process(
        self,
        request: Request,
        *,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        redirect_uri: Optional[str] = None,
        convert_response: Literal[True] = True,
    ) -> Optional[OpenID]: ...

    @overload
    async def verify_and_process(
        self,
        request: Request,
        *,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        redirect_uri: Optional[str] = None,
        convert_response: Literal[False],
    ) -> Optional[Dict[str, Any]]: ...

    async def verify_and_process(
        self,
        request: Request,
        *,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        redirect_uri: Optional[str] = None,
        convert_response: Union[Literal[True], Literal[False]] = True,
    ) -> Union[Optional[OpenID], Optional[Dict[str, Any]]]:
        """Processes the login given a FastAPI (Starlette) Request object. This should be used for the /callback path.

        Args:
            request (Request): FastAPI or Starlette request object.
            params (Optional[Dict[str, Any]]): Additional query parameters to pass to the provider.
            headers (Optional[Dict[str, Any]]): Additional headers to pass to the provider.
            redirect_uri (Optional[str]): Overrides the `redirect_uri` specified on this instance.
            convert_response (bool): If True, userinfo response is converted to OpenID object.

        Raises:
            SSOLoginError: If the 'code' parameter is not found in the callback request.

        Returns:
            Optional[OpenID]: User information as OpenID instance (if convert_response == True)
            Optional[Dict[str, Any]]: The original JSON response from the API.
        """
        headers = headers or {}
        code = unquote(request.query_params.get('code'))
        if code is None:
            logger.debug(
                'Callback request:\n\tURI: %s\n\tHeaders: %s\n\tQuery params: %s',
                request.url,
                request.headers,
                request.query_params,
            )
            raise SSOLoginError(400, "'code' parameter was not found in callback request")
        self._state = request.query_params.get('state')
        pkce_code_verifier: Optional[str] = None
        if self.uses_pkce:
            pkce_code_verifier = request.cookies.get('pkce_code_verifier')
            if pkce_code_verifier is None:
                warnings.warn(
                    'PKCE code verifier was not found in the request Cookie. This will probably lead to a login error.'
                )
        try:
            return await self.process_login(
                code,
                request,
                params=params,
                additional_headers=headers,
                redirect_uri=redirect_uri,
                pkce_code_verifier=pkce_code_verifier,
                convert_response=convert_response,
            )
        except Exception as e:
            raise SSOExceptionError(msg=str(e))

    def __enter__(self) -> 'SSOBase':
        self._oauth_client = None
        self._refresh_token = None
        self._id_token = None
        self._state = None
        if self.requires_state:
            self._generated_state = generate_random_state()
        if self.uses_pkce:
            self._pkce_code_verifier, self._pkce_code_challenge = get_pkce_challenge_pair(self._pkce_challenge_length)
        return self

    def __exit__(
        self,
        _exc_type: Optional[Type[BaseException]],
        _exc_val: Optional[BaseException],
        _exc_tb: Optional[TracebackType],
    ) -> None:
        return None

    @property
    def _extra_query_params(self) -> Dict:
        return {}

    @overload
    async def process_login(
        self,
        code: str,
        request: Request,
        *,
        params: Optional[Dict[str, Any]] = None,
        additional_headers: Optional[Dict[str, Any]] = None,
        redirect_uri: Optional[str] = None,
        pkce_code_verifier: Optional[str] = None,
        convert_response: Literal[True] = True,
    ) -> Optional[OpenID]: ...

    @overload
    async def process_login(
        self,
        code: str,
        request: Request,
        *,
        params: Optional[Dict[str, Any]] = None,
        additional_headers: Optional[Dict[str, Any]] = None,
        redirect_uri: Optional[str] = None,
        pkce_code_verifier: Optional[str] = None,
        convert_response: Literal[False],
    ) -> Optional[Dict[str, Any]]: ...

    async def process_login(
        self,
        code: str,
        request: Request,
        *,
        params: Optional[Dict[str, Any]] = None,
        additional_headers: Optional[Dict[str, Any]] = None,
        redirect_uri: Optional[str] = None,
        pkce_code_verifier: Optional[str] = None,
        convert_response: Union[Literal[True], Literal[False]] = True,
    ) -> Union[Optional[OpenID], Optional[Dict[str, Any]]]:
        """Processes login from the callback endpoint to verify the user and request user info endpoint.
        It's a lower-level method, typically, you should use `verify_and_process` instead.

        Args:
            code (str): The authorization code.
            request (Request): FastAPI or Starlette request object.
            params (Optional[Dict[str, Any]]): Additional query parameters to pass to the provider.
            additional_headers (Optional[Dict[str, Any]]): Additional headers to be added to all requests.
            redirect_uri (Optional[str]): Overrides the `redirect_uri` specified on this instance.
            pkce_code_verifier (Optional[str]): A PKCE code verifier sent to the server to verify the login request.
            convert_response (bool): If True, userinfo response is converted to OpenID object.

        Raises:
            ReusedOauthClientWarning: If the SSO object is reused, which is not safe and caused security issues.

        Returns:
            Optional[OpenID]: User information in OpenID format if the login was successful (convert_response == True).
            Optional[Dict[str, Any]]: Original userinfo API endpoint response.
        """
        if self._oauth_client is not None:  # pragma: no cover
            self._oauth_client = None
            self._refresh_token = None
            self._id_token = None
            warnings.warn(
                (
                    'Reusing the SSO object is not safe and caused a security issue in previous versions.'
                    "To make sure you don't see this warning, please use the SSO object as a context manager."
                ),
                ReusedOauthClientWarning,
            )
        params = params or {}
        params.update(self._extra_query_params)
        additional_headers = additional_headers or {}
        additional_headers.update(self.additional_headers or {})

        url = request.url

        if not self.allow_insecure_http and url.scheme != 'https':
            current_url = str(url).replace('http://', 'https://')
        else:
            current_url = str(url)

        current_path = f'{url.scheme}://{url.netloc}{url.path}'

        if pkce_code_verifier:
            params.update({'code_verifier': pkce_code_verifier})

        token_url, headers, body = self.oauth_client.prepare_token_request(
            await self.token_endpoint,
            authorization_response=current_url,
            redirect_url=redirect_uri or self.redirect_uri or current_path,
            code=code,
            **params,
        )  # type: ignore

        if token_url is None:  # pragma: no cover
            return None

        headers.update(additional_headers)

        auth = httpx.BasicAuth(self.client_id, self.client_secret)

        async with httpx.AsyncClient() as session:
            response = await session.post(token_url, headers=headers, content=body, auth=auth)
            content = response.json()
            self._refresh_token = content.get('refresh_token')
            self._id_token = content.get('id_token')
            self.oauth_client.parse_request_body_response(json.dumps(content))

            uri, headers, _ = self.oauth_client.add_token(await self.userinfo_endpoint)
            headers.update(additional_headers)
            session.headers.update(headers)
            response = await session.get(uri)
            content = response.json()
            if convert_response:
                return await self.openid_from_response(content, session)

            return content



from ... import settings
from ..sso.facebook import FacebookSSO


class FacebookService:
    def __init__(self):
        self.sso = FacebookSSO(
            client_id=settings.FACEBOOK_CLIENT_ID,
            client_secret=settings.FACEBOOK_CLIENT_SECRET,
            use_state=True,
            allow_insecure_http=True,
        )

    async def build_auth_url(self, redirect_uri: str):
        with self.sso:
            self.sso.redirect_uri = redirect_uri
            url = await self.sso.get_login_redirect(params={'prompt': 'consent', 'access_type': 'offline'})

        return url

    async def get_profile(self, request, redirect_uri: str):
        with self.sso:
            user = await self.sso.verify_and_process(request, redirect_uri=redirect_uri)

        return user



"""PKCE-related helper functions."""

import base64
import hashlib
import os
from typing import Tuple


def get_code_verifier(length: int = 96) -> str:
    """Get code verifier for PKCE challenge."""
    length = max(43, min(length, 128))
    bytes_length = int(length * 3 / 4)
    return base64.urlsafe_b64encode(os.urandom(bytes_length)).decode('utf-8').replace('=', '')[:length]


def get_pkce_challenge_pair(verifier_length: int = 96) -> Tuple[str, str]:
    """Get tuple of (verifier, challenge) for PKCE challenge."""
    code_verifier = get_code_verifier(verifier_length)
    code_challenge = (
        base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode('utf-8')).digest())
        .decode('utf-8')
        .replace('=', '')
    )

    return (code_verifier, code_challenge)



from pydantic import ConfigDict, BaseModel

from ..oauth.timezone import Timestamp


class AccessTokenBase(BaseModel):
    model_config = ConfigDict(from_attributes=True)
    access_token: str
    access_token_type: str = 'Bearer'
    access_token_expire_time: Timestamp


class GetLoginToken(AccessTokenBase):
    refresh_token: str
    refresh_token_type: str = 'Bearer'
    refresh_token_expire_time: Timestamp



from fastapi import APIRouter

from ..routers.auth import auth_router
from ..routers.user import user_router
from conf import settings
from ..routers.social import social_router

router = APIRouter(prefix=settings.API_V1_STR)
router.include_router(social_router)
router.include_router(auth_router)
router.include_router(user_router)



"""Twitter (X) SSO Oauth Helper class."""

from typing import TYPE_CHECKING, ClassVar, Optional

from .base import DiscoveryDocument, OpenID, SSOBase
from .constant import ProviderEnum

if TYPE_CHECKING:
    import httpx  # pragma: no cover


class TwitterSSO(SSOBase):
    """Class providing login via Twitter SSO."""

    provider = ProviderEnum.TWITTER
    scope: ClassVar = ['users.read', 'tweet.read']
    uses_pkce = True
    requires_state = True

    async def get_discovery_document(self) -> DiscoveryDocument:
        return {
            'authorization_endpoint': 'https://twitter.com/i/oauth2/authorize',
            'token_endpoint': 'https://api.twitter.com/2/oauth2/token',
            'userinfo_endpoint': 'https://api.twitter.com/2/users/me',
        }

    async def openid_from_response(self, response: dict, session: Optional['httpx.AsyncClient'] = None) -> OpenID:
        first_name, *last_name_parts = response['data'].get('name', '').split(' ')
        last_name = ' '.join(last_name_parts) if last_name_parts else None
        return OpenID(
            id=str(response['data']['id']),
            display_name=response['data']['username'],
            first_name=first_name,
            last_name=last_name,
            provider=self.provider,
        )



"""Fitbit OAuth Login Helper."""

from typing import TYPE_CHECKING, ClassVar, Optional

from .base import DiscoveryDocument, OpenID, SSOBase, SSOLoginError
from .constant import ProviderEnum

if TYPE_CHECKING:
    import httpx  # pragma: no cover


class FitbitSSO(SSOBase):
    """Class providing login via Fitbit OAuth."""

    provider = ProviderEnum.FITBIT
    scope: ClassVar = ['profile']

    async def openid_from_response(self, response: dict, session: Optional['httpx.AsyncClient'] = None) -> OpenID:
        """Return OpenID from user information provided by Google."""
        info = response.get('user')
        if not info:
            raise SSOLoginError(401, 'Failed to process login via Fitbit')
        return OpenID(
            id=info['encodedId'],
            first_name=info['fullName'],
            display_name=info['displayName'],
            picture=info['avatar'],
            provider=self.provider,
        )

    async def get_discovery_document(self) -> DiscoveryDocument:
        """Get document containing handy urls."""
        return {
            'authorization_endpoint': 'https://www.fitbit.com/oauth2/authorize?response_type=code',
            'token_endpoint': 'https://api.fitbit.com/oauth2/token',
            'userinfo_endpoint': 'https://api.fitbit.com/1/user/-/profile.json',
        }



#!/usr/bin/env python3
from pydantic import ConfigDict, BaseModel


class UserSchemaBase(BaseModel):
    model_config = ConfigDict(from_attributes=True)
    email: str | None = None
    picture: str | None = None
    first_name: str | None = None
    last_name: str | None = None
    phone: str | None = None


class RegisterUserParam(UserSchemaBase):
    pass


class UpdateUserParam(UserSchemaBase):
    pass


class CurrentUserInfo(UserSchemaBase):
    pass



import inspect
from typing import Optional

from fastapi import Request
from loguru import logger
from sqlalchemy import Integer, String, TIMESTAMP
from sqlalchemy.ext.asyncio import AsyncSession

from .. import settings
from ..cruds.crud_model_base import BaseDao
from ..oauth import jwt
from ..oauth.timezone import timezone
from ..social.services import *
from ..social.sso.base import OpenID
from ..social.sso.errors import SSOExceptionError
from ..cruds.crud_social_account import SocialAccountDao
from ..schemas.social_account import RegisterSocialAccountParam

from ..social.sso.constant import ProviderEnum


class SocialService:
    def __init__(self, provider):
        self.provider = provider
        self.sso_service = self.__init_sso__by_provider()

    def __init_sso__by_provider(self):
        match self.provider:
            case ProviderEnum.GOOGLE:
                return GoogleService()
            case ProviderEnum.GITHUB:
                return GithubService()
            case ProviderEnum.GITLAB:
                return GitlabService()
            case ProviderEnum.FACEBOOK:
                return FacebookService()
            case ProviderEnum.FITBIT:
                return FitbitService()
            case ProviderEnum.LINE:
                return LineService()
            case ProviderEnum.LINKEDIN:
                return LinkedInService()
            case ProviderEnum.NOTION:
                return NotionService()
            case ProviderEnum.TWITTER:
                return TwitterService()
            case _:
                raise SSOExceptionError(msg='Provider invalid')

    async def build_auth_url(self, redirect_uri: str):
        return await self.sso_service.build_auth_url(redirect_uri)

    async def verify_login(self, request: Request, redirect_uri: Optional[str] = None):
        try:
            # user_profile = await self.sso_service.sso.verify_and_process(request, redirect_uri=redirect_uri)
            usr_dict = {'id': '103127236837536345788', 'email': 'training.team113@gmail.com', 'first_name': 'training',
                        'last_name': 'team', 'display_name': 'training team',
                        'picture': 'https://lh3.googleusercontent.com/a/ACg8ocJ0wepnpXclYYMWK4kGDysMquVC4-FS3th5YRvK4qhEeC9hgQ=s96-c',
                        'provider': 'google'}
            user_profile = OpenID(**usr_dict)
            async with settings.DB_SESSION() as db:
                account = await SocialAccountDao.get_by_login_id_and_provider(
                    db=db, login_id=user_profile.id, provider=user_profile.provider
                )

                if account is None:
                    user, primary_key = await self.build_user_model_inst_from_open_id(user_profile)
                    user_model = await BaseDao.create_(db, create_data=user)
                    account = await self.create_new_account(
                        db, user_profile.id, str(user_model.__dict__.get(primary_key)), user_profile.provider
                    )

                access_token, access_token_expire_time = await jwt.create_access_token(
                    str(account.login_id), provider=user_profile.provider
                )
                refresh_token, refresh_token_expire_time = await jwt.create_refresh_token(
                    str(account.login_id), access_token_expire_time, provider=user_profile.provider
                )
                await db.commit()
                return (access_token, refresh_token, access_token_expire_time, refresh_token_expire_time)
        except Exception as e:
            logger.error(e)
            raise e

    async def create_new_account(self, db: AsyncSession, login_id, user_id, provider):
        register_account = RegisterSocialAccountParam(login_id=login_id, provider=provider, user_id=user_id)
        account = await SocialAccountDao.create(db, register_account)
        return account

    async def build_user_model_inst_from_open_id(self, user_profile: OpenID) -> settings.USER_MODEL:
        if settings.USER_MODEL is None:
            return None
        user_req = dict()
        columns = settings.USER_MODEL.__table__.columns
        constructor_params = [
            param for param in inspect.signature(settings.USER_MODEL.__init__).parameters.keys() if param != 'self'
        ]
        primary_key = None
        for column in columns:
            if column.primary_key:
                primary_key = column.name

            if column.name not in constructor_params:
                continue

            if column.default or (column.autoincrement and column.autoincrement not in ['auto', 'ignore_fk']):
                user_req.update({column.name: None})
                continue

            if not column.primary_key and column.nullable and not isinstance(column.type, String):
                user_req.update({column.name: None})
                continue

            if isinstance(column.type, Integer):
                user_req.update({column.name: 0})
            elif isinstance(column.type, TIMESTAMP):
                user_req.update({column.name: timezone.now})
            elif isinstance(column.type, String):
                if 'first_name' in column.name:
                    user_req.update({column.name: user_profile.first_name})
                elif 'last_name' in column.name:
                    user_req.update({column.name: user_profile.last_name})
                elif 'name' in column.name:
                    if user_profile.display_name:
                        user_req.update({column.name: user_profile.display_name})
                    else:
                        name = (
                            user_profile.first_name
                            if user_profile.first_name
                            else '' + user_profile.last_name
                            if user_profile.last_name
                            else ''
                        )
                        user_req.update({column.name: name})
                elif 'email' in column.name:
                    user_req.update({column.name: user_profile.email})

        return settings.USER_MODEL(**user_req), primary_key



from typing import Any

from pydantic import BaseModel
from sqlalchemy import inspect
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql.elements import Label
from sqlalchemy.sql.schema import Column

from .base import Base
import uuid


def _extract_matching_columns_from_schema(model: type[Base], schema: type[BaseModel] | list | None) -> list[Any]:
    """Retrieves a list of ORM column objects from a SQLAlchemy model that match the field names in a given
    Pydantic schema.

    Parameters
    ----------
    model: Type[Base]
        The SQLAlchemy ORM model containing columns to be matched with the schema fields.
    schema: Type[BaseModel]
        The Pydantic schema containing field names to be matched with the model's columns.

    Returns
    -------
    list[Any]
        A list of ORM column objects from the model that correspond to the field names defined in the schema.
    """
    column_list = list(model.__table__.columns)
    if schema is not None:
        if isinstance(schema, list):
            schema_fields = schema
        else:
            schema_fields = schema.model_fields.keys()

        column_list = [getattr(model, column_name) for column_name in schema_fields if hasattr(model, column_name)]
    return column_list


def _extract_matching_columns_from_kwargs(model: type[Base], kwargs: dict) -> list[Any]:
    if kwargs is not None:
        kwargs_fields = kwargs.keys()
        column_list = [getattr(model, column_name) for column_name in kwargs_fields if hasattr(model, column_name)]
    return column_list


def _extract_matching_columns_from_column_names(model: type[Base], column_names: list) -> list[Any]:
    return [getattr(model, column_name) for column_name in column_names if hasattr(model, column_name)]


def _auto_detect_join_condition(
    base_model: type[DeclarativeMeta], join_model: type[DeclarativeMeta]
) -> ColumnElement | None:
    """Automatically detects the join condition for SQLAlchemy models based on foreign key relationships. This
    function scans the foreign keys in the base model and tries to match them with columns in the join model.

    Parameters
    ----------
    base_model : Type[DeclarativeMeta]
        The base SQLAlchemy model from which to join.
    join_model : Type[DeclarativeMeta]
        The SQLAlchemy model to join with the base model.

    Returns
    -------
    Optional[ColumnElement]
        A SQLAlchemy ColumnElement representing the join condition, if successfully detected.

    Raises
    ------
    ValueError
        If the join condition cannot be automatically determined, a ValueError is raised.

    Example
    -------
    # Assuming User has a foreign key reference to Tier:
    join_condition = auto_detect_join_condition(User, Tier)
    """
    fk_columns = [col for col in inspect(base_model).c if col.foreign_keys]
    join_on = next(
        (
            base_model.__table__.c[col.name] == join_model.__table__.c[list(col.foreign_keys)[0].column.name]
            for col in fk_columns
            if list(col.foreign_keys)[0].column.table == join_model.__table__
        ),
        None,
    )

    if join_on is None:
        raise ValueError('Could not automatically determine join condition. Please provide join_on.')

    return join_on


def _add_column_with_prefix(column: Column, prefix: str | None) -> Label:
    """Creates a SQLAlchemy column label with an optional prefix.

    Parameters
    ----------
    column : Column
        The SQLAlchemy Column object to be labeled.
    prefix : Optional[str]
        An optional prefix to prepend to the column's name.

    Returns
    -------
    Label
        A labeled SQLAlchemy Column object.
    """
    column_label = f'{prefix}{column.name}' if prefix else column.name
    return column.label(column_label)


def generate_uuid() -> str:
    return str(uuid.uuid4())



#!/usr/bin/env python3
from datetime import datetime
from typing import Any

from pydantic import BaseModel, ConfigDict

from ..conf import settings

_ExcludeData = set[int | str] | dict[int | str, Any]

__all__ = ['ResponseModel', 'response_base']


class ResponseModel(BaseModel):
    """
    Unified return model

    .. tip::

        If you don't want to use the custom encoder in ResponseBase, you can use this model,
        and the return data will be automatically parsed and returned through the encoder inside fastapi;
        This return model generates openapi schema documentation

    E.g. ::

        @router.get('/test', response_model=ResponseModel)
        def test():
            return ResponseModel(data={'test': 'test'})

        @router.get('/test')
        def test() -> ResponseModel:
            return ResponseModel(data={'test': 'test'})

        @router.get('/test')
        def test() -> ResponseModel:
            res = CustomResponseCode.HTTP_200
            return ResponseModel(code=res.code, msg=res.msg, data={'test': 'test'})
    """

    # TODO: json_encoders Configuration failure: https://github.com/tiangolo/fastapi/discussions/10252
    model_config = ConfigDict(json_encoders={datetime: lambda x: x.strftime(settings.DATETIME_FORMAT)})

    code: int = 200
    msg: str = 'Request successful'
    data: Any | None = None


class ResponseBase:
    """
    Unified return method

    .. tip::

        The return method in this class will return the ResponseModel model, which exists as a coding style;

    E.g. ::

        @router.get('/test')
        def test() -> ResponseModel:
            return await response_base.success(data={'test': 'test'})
    """

    @staticmethod
    async def __response(*, code, msg, data: Any | None = None) -> ResponseModel:
        """
        The request returns a common method if successful

        :param res: Return messages
        :param data: Return data
        :return:
        """
        return ResponseModel(code=code, msg=msg, data=data)

    async def success(
        self,
        *,
        code=200,
        msg='Request successful',
        data: Any | None = None,
    ) -> ResponseModel:
        return await self.__response(code=code, msg=msg, data=data)

    async def fail(
        self,
        *,
        code=400,
        msg='Bad Request',
        data: Any = None,
    ) -> ResponseModel:
        return await self.__response(code=code, msg=msg, data=data)


response_base = ResponseBase()



"""Notion SSO Oauth Helper class."""

from typing import TYPE_CHECKING, ClassVar, Optional

from .base import DiscoveryDocument, OpenID, SSOBase, SSOLoginError
from .constant import ProviderEnum

if TYPE_CHECKING:
    import httpx  # pragma: no cover


class NotionSSO(SSOBase):
    """Class providing login using Notion OAuth."""

    provider = ProviderEnum.NOTION
    scope: ClassVar = ['openid']
    additional_headers: ClassVar = {'Notion-Version': '2022-06-28'}

    async def get_discovery_document(self) -> DiscoveryDocument:
        return {
            'authorization_endpoint': 'https://api.notion.com/v1/oauth/authorize?owner=user',
            'token_endpoint': 'https://api.notion.com/v1/oauth/token',
            'userinfo_endpoint': 'https://api.notion.com/v1/users/me',
        }

    async def openid_from_response(self, response: dict, session: Optional['httpx.AsyncClient'] = None) -> OpenID:
        owner = response['bot']['owner']
        if owner['type'] != 'user':
            raise SSOLoginError(401, f"Notion login failed, owner is not a user but {response['bot']['owner']['type']}")
        return OpenID(
            id=owner['user']['id'],
            email=owner['user']['person']['email'],
            picture=owner['user']['avatar_url'],
            display_name=owner['user']['name'],
            provider=self.provider,
        )



#!/usr/bin/env python3

from fastapi import APIRouter

from ..schemas.response_schema import response_base
from ..schemas.social_account import RegisterAccountParam
from ..services.user import UserService

user_router = APIRouter(prefix='/users', tags=['Users'])


@user_router.post(path='/register', summary='Registration')
async def register_user(obj: RegisterAccountParam):
    data = await UserService().register_user(obj=obj)
    return await response_base.success(data=data)



#!/usr/bin/env python3
import uuid

from sqlalchemy import select, and_
from sqlalchemy.ext.asyncio import AsyncSession

from ..models.t_social_account import SocialAccount
from ..oauth.jwt import get_hash_password
from ..schemas.social_account import RegisterSocialAccountParam, UpdateSocialAccountParam

from .base import CRUDBase


class CRUDSocialAccount(CRUDBase[SocialAccount, RegisterSocialAccountParam, UpdateSocialAccountParam]):
    async def get_by_login_id_and_provider(
        self, db: AsyncSession, login_id: str, provider: str
    ) -> SocialAccount | None:
        user = await db.execute(
            select(self.model).where(and_(self.model.login_id == login_id, self.model.provider == provider))
        )
        return user.scalars().first()

    async def get_by_user_id_and_provider(self, db: AsyncSession, user_id: int, provider: str) -> SocialAccount | None:
        user = await db.execute(
            select(self.model).where(and_(self.model.user_id == user_id, self.model.provider == provider))
        )
        return user.scalars().first()

    async def create(self, db: AsyncSession, obj: RegisterSocialAccountParam, **kwargs) -> SocialAccount:
        if obj.password:
            obj.password = await get_hash_password(obj.password)
        else:
            obj.password = await get_hash_password(uuid.uuid4().hex)

        dict_obj = obj.model_dump()
        new_user = self.model(**dict_obj)
        db.add(new_user)
        await db.flush()
        return new_user


SocialAccountDao: CRUDSocialAccount = CRUDSocialAccount(SocialAccount)



from enum import Enum


class ProviderEnum(str, Enum):
    FACEBOOK: str = 'facebook'
    GITHUB: str = 'github'
    GOOGLE: str = 'google'
    FITBIT: str = 'fitbit'
    GITLAB: str = 'gitlab'
    LINE: str = 'line'
    LINKEDIN: str = 'linkedin'
    NOTION: str = 'notion'
    TWITTER: str = 'twitter'



#!/usr/bin/env python3
from typing import Type, TypeVar

from sqlalchemy.ext.asyncio import AsyncSession

from ..models.base import MappedBase

ModelType = TypeVar('ModelType', bound=MappedBase)


class CRUDModelBase:
    """Base class for CRUD operations on a model.

    Parameters
    ----------
    model : Type[ModelType]
        The SQLAlchemy model type.
    """

    async def create_(self, db: AsyncSession, create_data: ModelType) -> ModelType:
        """Create a new record in the database.

        :param db: AsyncSession
        :param create_data:  Model class
        :return:
        """
        db.add(create_data)
        await db.flush()
        return create_data


BaseDao: CRUDModelBase = CRUDModelBase()



#!/usr/bin/env python3
from typing import Optional, Any

from pydantic import ConfigDict, BaseModel

from ..schemas.user import RegisterUserParam


class SocialAccountSchemaBase(BaseModel):
    model_config = ConfigDict(from_attributes=True)
    login_id: str
    provider: Optional[str] = None
    password: Optional[str] = None
    user_id: Optional[Any] = None


class RegisterSocialAccountParam(SocialAccountSchemaBase):
    pass


class UpdateSocialAccountParam(SocialAccountSchemaBase):
    pass


class RegisterAccountParam(SocialAccountSchemaBase, RegisterUserParam):
    pass
